package me.hammerle.math;

import java.util.TreeSet;

public final class Fraction implements Comparable<Fraction>
{
    private final long numerator;
    private final long denominator;
    
    public Fraction(long numerator, long denominator)
    {
        if(denominator == 0)
        {
            throw new ArithmeticException();
        }
        
        if(denominator != 1)
        {
            long divisor = getGreatestCommonDivisor(numerator, denominator);
            if(divisor != 1)
            {
                denominator /= divisor;
                numerator /= divisor;
            }
        }
        
        // short fraction
        if(denominator < 0)
        {
            this.denominator = -denominator;
            this.numerator = -numerator;
        }
        else
        {
            this.denominator = denominator;
            this.numerator = numerator;
        }
    }
    
    public Fraction(long numerator)
    {
        this(numerator, 1);
    }
    
    // -----------------------------------------------------------------------------------
    // constants, chain fractions
    // ----------------------------------------------------------------------------------- 
    
    /*public static final Fraction PI = Fraction.getChainFraction(
              //3,7,15,1,292,1,1,1,2,1,3,1,14,2,1,1,2,2,2,2,1,84,2,1,1,15,3,13,1,4,2,6,6
                3,7,15,1,292,1,1,1,2,1,3,1,14,2,1);
    public static final Fraction E = Fraction.getChainFraction(
              //2,1,2,1,1,4,1,1,6,1,1,8,1,1,10,1,1,12,1,1,14,1,1,16,...
                2,1,2,1,1,4,1,1,6,1,1,8,1,1,10,1,1,12,1,1,14);*/
    
    public static Fraction getChainFraction(int... ints)
    {
        if(ints.length == 1)
        {
            return new Fraction(ints[0]);
        }
        return new Fraction(ints[0]).add(getChainFractionPart(1, ints));
    }
    
    private static Fraction getChainFractionPart(int index, int... ints)
    {
        if(index + 1 == ints.length)
        {
            return new Fraction(1, ints[index]);
        }
        return new Fraction(ints[index]).add(getChainFractionPart(index + 1, ints)).invert();
    }
    
    // -----------------------------------------------------------------------------------
    // basic calculating
    // ----------------------------------------------------------------------------------- 
    
    public Fraction add(Fraction f)
    {
        if(denominator == f.denominator)
        {
            return new Fraction(Math.addExact(numerator, f.numerator), denominator);
        }
        else
        {
            long l = getLeastCommonMultiple(denominator, f.denominator);
            return new Fraction(Math.addExact(Math.multiplyExact(numerator, l / denominator),
                    Math.multiplyExact(f.numerator, l / f.denominator)), l);
        }
    }
    
    public Fraction mul(Fraction f)
    {
        return new Fraction(Math.multiplyExact(numerator, f.numerator), 
                Math.multiplyExact(denominator, f.denominator));
    }
    
    public Fraction sub(Fraction f)
    {
        return add(f.invertSign());
    }
    
    public Fraction div(Fraction f)
    {
        return mul(f.invert());
    }
    
    // -----------------------------------------------------------------------------------
    // roots, power
    // ----------------------------------------------------------------------------------- 
    
    private static long power(long l, long power)
    {
        long factor = 1;
        if(l < 0)
        {
            factor = (power & 1) == 1 ? -1 : 1;
            l = -l;
        }
        long prod = 1;
        while(power > 0)
        {
            if((power & 1) == 1)
            {
                prod = Math.multiplyExact(prod, l);
            }
            power = power >> 1;
            l = Math.multiplyExact(l, l);
        }
        if(factor == -1)
        {
            return -prod;
        }
        return prod;
    }
    
    private static long rootOfLong(long l, long root)
    {
        if(l == 0 || l == 1)
        {
            return l;
        }
        try
        {
            TreeSet<Long> tree = new TreeSet<>();
            long currentValue = l >> 1; // taking half as start value
            long an = currentValue;
            do
            {
                tree.add(currentValue);
                currentValue = currentValue - (currentValue / root) + an / power(currentValue, root - 1); 
            }
            while(!tree.contains(currentValue));
            return currentValue;
        }
        catch(ArithmeticException ex)
        {
            return 0;
        } 
    }
    
    private Fraction root(long root)
    {
        if(root == 1)
        {
            return this.copy();
        }
        // Getting nice start value
        Fraction currentValue;
        Fraction n = new Fraction(root);
        Fraction an = this.div(n);

        Fraction newFraction = new Fraction(rootOfLong(numerator, root), rootOfLong(denominator, root));
        root--;
        try
        {
            do
            {
                currentValue = newFraction;
                newFraction = currentValue.sub(currentValue.div(n)).add(an.div(currentValue.power(root)));              
            }
            while(!newFraction.equals(currentValue));
        }
        catch(ArithmeticException ex)
        {
        }
        return newFraction;
    }
    
    public Fraction power(Fraction f)
    {
        if(f.numerator == 0)
        {
            return new Fraction(1);
        }
        return this.power(f.numerator).root(f.denominator);
    }
    
    private Fraction power(long p)
    {
        if(p < 0)
        {
            p = -p;
            invert();
        }
        else if(p == 1)
        {
            return this.copy();
        }
        long prodn = 1;
        long prodd = 1;
        long n = numerator;
        long d = denominator;
        while(p > 0)
        {
            if((p & 1) == 1)
            {
                prodn = Math.multiplyExact(prodn, n);
                prodd = Math.multiplyExact(prodd, d);
            }
            p = p >> 1;
            n = Math.multiplyExact(n, n);
            d = Math.multiplyExact(d, d);
        }
        return new Fraction(prodn, prodd);
    }
    
    // -----------------------------------------------------------------------------------
    // inverting
    // ----------------------------------------------------------------------------------- 
    
    public Fraction invertSign()
    {
        return new Fraction(-numerator, denominator);
    }
    
    public Fraction invert()
    {
        if(numerator == 0)
        {
            throw new ArithmeticException();
        }
        else if(numerator < 0)
        {
            return new Fraction(-denominator, -numerator);
        }
        return new Fraction(denominator, numerator);
    }
    
    // -----------------------------------------------------------------------------------
    // trigonometric
    // ----------------------------------------------------------------------------------- 
    
    public Fraction sin()
    {
        /*int factor = 1;
        if(numerator < 0)
        {
            factor = -1;
            numerator = -numerator;
        }
        
        Fraction pi = Fraction.PI.mul(2);
        while(this.compareTo(pi) >= 0)
        {
            this.sub(pi);
        }
        
        if(isBetween(new Fraction(0), Fraction.PI.div(2)))
        {
            // Nichts
        }
        else if(isBetween(Fraction.PI.div(2), Fraction.PI))
        {
            this.invertSign().add(Fraction.PI);
        }
        else if(this.compareTo(Fraction.PI) >= 0)
        {
            this.sub(Fraction.PI);
            factor = -factor;
        }
        
        
        Fraction quad = this.clone().power(2);
        Fraction value = this.clone();
        Fraction result = this.clone();
        
        int k = 1;
        for(int i = 0; i < 20; i++)
        {
            value.div(i + 2);
            if((i & 1) == 1)
            {
                value.mul(quad);
                k = -k;
                if(k == -1)
                {
                    result.sub(value);
                }
                else
                {
                    result.add(value);
                }
            }
        }
        numerator = factor * result.numerator;
        denominator = result.denominator;*/
        return this;
    }
    
    public Fraction cos()
    {
        return this;
        //return add(Fraction.PI.div(new Fraction(2))).sin();
    }
    
    public Fraction tan()
    {
        return sin().div(cos());
    }
    
    // -----------------------------------------------------------------------------------
    // simplifying
    // ----------------------------------------------------------------------------------- 
    
    /*private void simplify()
    {
        if(denominator == 1)
        {
            throw new ArithmeticException();
        }
        denominator = (denominator & 1) == 1 ? denominator - 1 : denominator;
        numerator = (numerator & 1) == 1 ? numerator - 1 : numerator;
        shortFraction();
    }*/
    
    private long getGreatestCommonDivisor(long i, long n)
    {
        if(i == 0)
        {
            return n;
        }
        if(i < 0)
        {
            i = -i;
        }
        if(n < 0)
        {
            n = -n;
        }
        long helper;
        while(true)
        {
            if(i < n)
            {
                helper = i;
                i = n;
                n = helper;
            }
            i = i % n;
            if(i == 0)
            {
                return n;
            }
        }
    }
    
    private long getLeastCommonMultiple(long i, long n)
    {
        return Math.abs(Math.multiplyExact(i, n)) / getGreatestCommonDivisor(i, n);
    }
    
    // -----------------------------------------------------------------------------------
    // basic stuff
    // ----------------------------------------------------------------------------------- 
    
    public double getDouble()
    {
        return numerator / (double) denominator;
    }
    
    public long getLong()
    {
        return Math.round(numerator / (double) denominator);
    }
    
    @Override
    public String toString() 
    {
        if(denominator == 1)
        {
            return String.valueOf(numerator);
        }
        return numerator + " / " + denominator;
    }

    public Fraction copy()
    {
        return new Fraction(numerator, denominator);
    }

    @Override
    public boolean equals(Object o) 
    {
        if(o == null)
        {
            return false;
        }
        else if(o.getClass() != Fraction.class)
        {
            return false;
        }
        Fraction f = (Fraction) o;
        return numerator == f.numerator && denominator == f.denominator;
    }

    @Override
    public int hashCode() 
    {
        int hash = 3;
        hash = 97 * hash + (int) (this.numerator ^ (this.numerator >>> 32));
        hash = 97 * hash + (int) (this.denominator ^ (this.denominator >>> 32));
        return hash;
    }

    @Override
    public int compareTo(Fraction f) 
    {
        if(f.numerator < 0 && numerator >= 0)
        {
            return 1;
        }
        else if(f.numerator >= 0 && numerator < 0)
        {
            return -1;
        }
        else
        {
            long i = f.sub(this).numerator;
            if(i == 0)
            {
                return 0;
            }
            else if(i < 0)
            {
                return 1;
            }
            else
            {
                return -1;
            }
        }
    }
    
    public boolean isBetween(Fraction f1, Fraction f2)
    {
        return this.compareTo(f1) >= 0 && this.compareTo(f2) <= 0;
    }
}
